package meerkat.bulletinboard.sqlserver;

import java.util.Arrays;
import java.util.List;

import javax.ws.rs.GET;
import javax.ws.rs.POST;
import javax.ws.rs.Path;

import com.google.protobuf.ProtocolStringList;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;

import meerkat.bulletinboard.BulletinBoardServer;
import meerkat.protobuf.Voting.MessageFilterList;
import meerkat.comm.CommunicationException;
import meerkat.protobuf.Crypto.SignatureVerificationKey;
import meerkat.protobuf.Voting.BulletinBoardMessage;
import meerkat.protobuf.Voting.BulletinBoardMessageList;
import meerkat.crypto.Digest;
import meerkat.crypto.concrete.SHA256Digest;
import meerkat.protobuf.Voting.MessageID;

public abstract class BulletinBoardSQLServer implements BulletinBoardServer{
	
	protected Connection connection;
	
	protected Digest digest;
	
	protected List<SignatureVerificationKey> trusteeSignatureVerificationArray;
	protected int minTrusteeSignatures;

	protected List<List<SignatureVerificationKey>> pollingCommitteeSignatureVerificationKeyArray;
	protected int minCommiteeSignatures;
	
	/**
	 * This method initializes the signatures but does not implement the DB connection.
	 * Any full (non-abstract) extension of this class should
	 * 1. Establish a DB connection, and
	 * 2. Call this procedure
	 */
	@Override
	public void init() throws CommunicationException {
		// TODO write signature reading part.
		
		digest = new SHA256Digest();
	}

	/**
	 * This method verifies the authenticity of the received message based on
	 * the stored signatures.
	 * 
	 * @param msg
	 *            is the message to authenticate (containing the signature).
	 * @return TRUE if the message is authenticated and FALSE otherwise.
	 */
	private boolean verifyMessage(BulletinBoardMessage msg) {
		//TODO: Replace with actual verification.
		return true;
	}
	
	/**
	 * This procedure makes sure that all tags in the given list have an entry in the tags list. 
	 * @param tagIterator
	 */
	protected abstract void insertNewTags(String[] tags) throws SQLException;
	
	@Override
	public boolean postMessage(BulletinBoardMessage msg) throws CommunicationException {
		if (!verifyMessage(msg)) {
			return false;
		}
		
		PreparedStatement pstmt;
		String sql;
		byte[] msgID;
		
		ProtocolStringList tagList;
		String[] tags;
		
		digest.reset();
		digest.update(msg);
		
		msgID = digest.digest();
		
		try {
			
			sql = "INSERT INTO MsgTable (MsgId, Msg, SignerId) VALUES(?,?,?)";
		    pstmt = connection.prepareStatement(sql);
		    pstmt.setBytes(1, msgID);
		    pstmt.setBytes(2, msg.toByteArray());
		    pstmt.setBytes(3, msg.getSig().getSignerId().toByteArray());
		    pstmt.executeUpdate();
		    pstmt.close();
		    
		} catch (SQLException e) {
			throw new CommunicationException("Error inserting into MsgTable: " + e.getMessage());
		}
		
		try {
			
		    tagList = msg.getMsg().getTagsList();
		    tags = new String[tagList.size()];
		    tags = tagList.toArray(tags);
		    
		    insertNewTags(tags);
		    
		} catch (SQLException e) {
			throw new CommunicationException(e.getMessage());
		}
		
		try{
		    sql = "INSERT INTO MsgTagTable (TagId, MsgId) SELECT TagTable.TagId, ? AS MsgId FROM TagTable WHERE Tag = ?";
	    	pstmt = connection.prepareStatement(sql);
		    
		    for (String tag : tags){
		    	pstmt.setBytes(1, msgID);
		    	pstmt.setString(2, tag);
		    	pstmt.addBatch();
		    }
		    
		    pstmt.executeBatch();
		    pstmt.close();
			
		} catch (SQLException e) {
			throw new CommunicationException("Error Linking tags: " + e.getMessage());
		}

		return true;
	}

	@Override
	public BulletinBoardMessageList readMessages(MessageFilterList filterList) {
		// TODO Auto-generated method stub
		return null;
	}
	
	public String testPrint(){
		
		String s = "";
		
		try {
	    
		    Statement statement = connection.createStatement();
			ResultSet rs = statement.executeQuery("select * from MsgTable");
			while (rs.next()) {
				// read the result set
				s += "entry = " + rs.getInt("EntryNum") + " \n";
				s += "id = " + Arrays.toString(rs.getBytes("MsgId")) + " \n";
				s += "msg = " +  Arrays.toString(rs.getBytes("Msg")) + " \n";
				s += "signer ID = " +  Arrays.toString(rs.getBytes("SignerId")) + "\t\n<BR>";
			}
			
			rs = statement.executeQuery("select * from TagTable");
			while (rs.next()) {
				// read the result set
				s += "Tag = " + rs.getString("Tag") + " \n";
				s += "TagId = " + rs.getInt("TagId") + "\t\n<BR>";
			}
			
			rs = statement.executeQuery("select * from MsgTagTable");
			while (rs.next()) {
				// read the result set
				s += "MsgId = " +  Arrays.toString(rs.getBytes("MsgId")) + " \n";
				s += "TagId = " + rs.getInt("TagId") + "\t\n<BR>";
			}
		} catch(SQLException e){
			s += "Error reading from DB";
		}
		
		return s;
	}

}
